import numpy as np
from numpy import pi
from scipy.special import zeta
import scipy.optimize as optimize


from numba import jit,njit,prange

#Credit Teresa Karanikolaou who extracted the expansions for Bethe Ansatz solution
#of the LL model at fixed density. It's much faster than solving directly the Bethe
#equation. As a consequence however, for intermediate values of the coupling, the results
#are not exact.

def energy_ristivojevic(gamma):
    """
    Implements equation 1 of 'Conjectures about the ground-state energy of the Lieb-Liniger model at weak repulsion' by Zoran Ristivojevic
    https://arxiv.org/pdf/1905.13705.pdf
    """
    dest=0.;
    dest+=gamma
    dest+=-4./(3.*pi)*gamma**(3/2)
    dest+=(pi**2-6)/(6*pi**2)*gamma**2
    dest+=-(4.-3*zeta(3))/(8*pi**3)*gamma**(5/2)
    dest+=-(4.-3*zeta(3))/(24*pi**4)*gamma**3
    dest+=-(45.*zeta(5)-60*zeta(3)+32)/(1024*pi**5)*gamma**(7/2)
    dest+=-3*(15*zeta(5)-4*zeta(3)-6*zeta(3)**2)/(2048*pi**6)*gamma**4
    dest+=-(8505*zeta(7)-2520*zeta(5)+4368*zeta(3)-6048*zeta(3)**2-1024)/(786432*pi**7)*gamma**(9/2)
    dest+=-9*(273*zeta(7)-120*zeta(5)+16*zeta(3)-120*zeta(3)*zeta(5))/(131072*pi**8)*gamma**5
    return dest

def energy_lang(gamma):
    """ Implementation of Appendix E of 'Ground-state energy and excitation spectrum of the Lieb-Liniger
    model: accurate analytical results and conjectures about the exact
    solution'
    https://www.scipost.org/SciPostPhys.3.1.003/pdf
    """
    #This expansion is only conjectured, not proven
    dest=0.
    dest+=1.
    dest+=-4./gamma
    dest+= 12/gamma**2
    dest+=-10.9448/gamma**3
    dest+=-130.552/gamma**4
    dest+= 804.13/gamma**5
    dest+=-910.345/gamma**6
    dest+=-15423.8/gamma**7
    dest+= 100559.0/gamma**8
    dest+=-67110.5/gamma**9
    dest+=-2.64681e6/gamma**10
    dest+= 1.552627e7/gamma**11
    dest+= 4.69185e6/gamma**12
    dest+=-5.35057e8/gamma**13
    dest+= 2.6096e9/gamma**14
    dest+= 4.84076e9/gamma**15
    dest+=-1.16548e11/gamma**16
    dest+= 4.35667e11/gamma**17
    dest+= 1.93421e12/gamma**18
    dest+=-2.60894e13/gamma**19
    dest+= 6.51416e13/gamma**20
    #Multiply with TG energy
    dest*=pi**2/3
    return dest

#Gives the Bethe-Ansatz energy for LL with a chemical potential
def energy_fixed_density(c):
    if c < 10:
        energy=energy_ristivojevic(c)
    else:
        energy=energy_lang(c)
    return energy

def energy_ll(c,mu):

    f= lambda x: x**3*energy_fixed_density(c/x)-mu*x
    result=optimize.minimize(f,1e-5)

    if  c/result.x < 20 and c/result.x >1:
        print("Waring: the coupling/chemical potential is in a range")
        print("where the series expansion could be inaccurate")
    print(result.fun)
