import numpy as np
import scipy.optimize as optimize
from scipy.sparse.linalg import LinearOperator,eigs



def dagger(M):
    return np.transpose(np.conjugate(M))

def Lindblad(R,K,rho):
    dR=np.transpose(np.conjugate(R))
    Q= -1j*K - 0.5*dR @ R
    dQ=np.transpose(np.conjugate(Q))
    return Q @ rho + rho @ dQ + R @ rho @ dR

def get_rho0(R,K):
    D=len(R)
    Lindblad_vector = lambda x : Lindblad(R,K,x.reshape(D,D)).reshape(D*D)
    rho_init=np.eye(D,dtype=complex)/D
    Lops= LinearOperator((D*D,D*D), matvec=Lindblad_vector,dtype=complex)
    vals, vecs = eigs(Lops,which="LR", k=1,v0=rho_init.reshape(D*D))
    rho0=vecs[:,0].reshape(D,D)
    rho0=rho0/np.trace(rho0)
    rho0=rho0*0.5+dagger(rho0)*0.5
    return rho0

def density(R,K,rho0):
    return np.trace(dagger(R) @ R @ rho0)

def interaction(R,K,rho0):
    return np.trace(dagger(R) @ dagger(R) @ R @ R @ rho0)

def kinetic(R,K,rho0):
    Q= -1j*K - 0.5*dagger(R) @ R
    comm=(Q @ R) - (R @ Q)
    return np.trace(dagger(comm) @ comm @ rho0)

def total_energy(x,D,c,mu):
    R,K=RK_from_parameters(x,D)
    rho0=get_rho0(R,K)
    energy=kinetic(R,K,rho0) + c*interaction(R,K,rho0)-mu* density(R,K,rho0)
    return energy.real

def RK_from_parameters(paramvector,D):
    R=np.zeros([D,D])*(1+1j)
    Rre=1.0*paramvector[:D**2]
    Rim=1.0*paramvector[D**2:2*D**2]
    R=Rre.reshape(D,D)+Rim.reshape(D,D)*1j
    Kdiag=paramvector[2*D**2:2*D**2+D]
    K=np.diag(Kdiag)
    return R,K

def callback(x,D,c,mu):
    #Just to get the energy printed at each iteration of
    #the black box optimizer
    print(total_energy(x,D,c,mu))
    return False

def naive_optimizer(c,mu,D):
    x0=np.random.normal(0,1,2*D*D+D)
    energy=lambda x : total_energy(x,D,c,mu)
    optimizer_callback = lambda x: callback(x,D,c,mu)
    result=optimize.minimize(energy,x0,callback=optimizer_callback,method="BFGS")
    print("Energy value "+str(result.fun))
    print("Optimizer status " + str(result.status) + "     (0=success, otherwise failure)")
