import numpy as np
from scipy.sparse.linalg import LinearOperator,eigs,bicg,cg,cgs,bicgstab


def generate_RK(D):
    x=np.random.normal(0,1,2*D*D+D)
    R,K=RK_from_parameters(x,D)
    return R,K


def RK_from_parameters(paramvector,D):
    R=np.zeros([D,D])*(1+1j)
    Rre=1.0*paramvector[:D**2]
    Rim=1.0*paramvector[D**2:2*D**2]
    R=Rre.reshape(D,D)+Rim.reshape(D,D)*1j
    Kdiag=paramvector[2*D**2:2*D**2+D]
    K=np.diag(Kdiag)
    return R,K

def comm(A,B):
    return (A @ B)-(B @ A)

def dagger(M):
    return np.transpose(np.conjugate(M))

def Lindblad(R,K,rho):
    dR=np.transpose(np.conjugate(R))
    Q= -1j*K - 0.5*dR @ R
    dQ=np.transpose(np.conjugate(Q))
    return Q @ rho + rho @ dQ + R @ rho @ dR

def Lindblad_adjoint(R,K,rho):
    dR=np.transpose(np.conjugate(R))
    Q= +1j*K - 0.5*dR @ R
    dQ=np.transpose(np.conjugate(Q))
    return Q @ rho + rho @ dQ + dR @ rho @ R

def get_rho0(R,K):
    D=len(R)
    Lindblad_vector = lambda x : Lindblad(R,K,x.reshape(D,D)).reshape(D*D)
    rho_init=np.eye(D,dtype=complex)/D
    Lops= LinearOperator((D*D,D*D), matvec=Lindblad_vector,dtype=complex)
    vals, vecs = eigs(Lops,which="LR", k=1,v0=rho_init.reshape(D*D))
    rho0=vecs[:,0].reshape(D,D)
    rho0=rho0/np.trace(rho0)
    rho0=rho0*0.5+dagger(rho0)*0.5
    return rho0

def density(R,K,rho0):
    return np.trace(dagger(R) @ R @ rho0)

def interaction(R,K,rho0):
    return np.trace(dagger(R) @ dagger(R) @ R @ R @ rho0)

def kinetic(R,K,rho0):
    Q= -1j*K - 0.5*dagger(R) @ R
    comm=(Q @ R) - (R @ Q)
    return np.trace(dagger(comm) @ comm @ rho0)

def total_energ(R,K,c,mu):
    rho0=get_rho0(R,K)
    energy=kinetic(R,K,rho0) + c*interaction(R,K,rho0)-mu* density(R,K,rho0)
    return energy.real



def Lh(R,K,c,mu):
    rho0=get_rho0(R,K)
    D=len(R)
    Q= -1j*K - 0.5*dagger(R) @ R
    energy_matrix = (- mu * dagger(R) @ R + c *dagger(R) @ dagger(R) @ R @ R
                    + dagger(comm(Q,R)) @ comm(Q,R))
    Lindblad_vector = lambda x : Lindblad(R,K,x.reshape(D,D)).reshape(D*D) + np.trace(x.reshape(D,D))*rho0.reshape(D*D)
    Lindblad_vector_adjoint = lambda x : Lindblad_adjoint(R,K,x.reshape(D,D)).reshape(D*D) + np.trace(x.reshape(D,D) @ rho0)*np.eye(D).reshape(D*D)

    Lops= LinearOperator((D*D,D*D), matvec=Lindblad_vector_adjoint,rmatvec=Lindblad_vector,dtype=complex)
    Lh,exitcode=bicg(Lops, energy_matrix.reshape(D*D))

    Lh=Lh.reshape(D,D)
    Lh=Lh-np.trace(energy_matrix @ rho0)*np.eye(len(R))
    Lh=-Lh #in fact, in Haegeman et al, Lh is defined from minus the transfer matrix

    return Lh

def get_W(R,K,c,mu):
    lh= Lh(R,K,c,mu)
    Q= -1j*K - 0.5*dagger(R) @ R
    rho0=get_rho0(R,K)
    Wrho=(dagger(Q) @ comm(Q,R)@rho0 - comm(Q,R) @ rho0 @ dagger(Q)
            - R @ comm(Q,R) @ rho0 @ dagger(R) + R @ dagger(R) @ comm(Q,R) @ rho0
            - mu*R @ rho0 + c *R @ R @rho0 @ dagger(R)
            +c *dagger(R) @ R @ R @ rho0
            +lh @ R @ rho0 - R @ lh @ rho0)
    W=Wrho @ np.linalg.inv(rho0)
    return W



def tdvp_optimizer(c,mu,D,maxit=10000,dt=1e-5):
    old_energy=1e10
    new_energy=1e10
    W=np.eye(D)
    nit=0
    R,K=generate_RK(D)
    while np.trace(dagger(W) @ W)/D**2 >1e-8 and nit<maxit:
        nit=nit+1
        dt=dt*2
        W=get_W(R,K,c,mu)
        Rbest=R-dt*W
        Kbest=K-dt*1j*0.5*(dagger(W) @ R - dagger(R) @ W)
        new_energy=total_energ(Rbest,Kbest,c,mu)

        temporary_dt=dt
        counter=0
        while new_energy > old_energy or counter < 2:
            #we reduce the time step until the energy is going down
            #this is very crude
            temporary_dt=temporary_dt/2
            Rnew=R-temporary_dt*W
            Knew=K-temporary_dt*1j*0.5*(dagger(W) @ R - dagger(R) @ W)
            temporary_energy=total_energ(Rnew,Knew,c,mu)
            counter=counter+1
            if temporary_energy < new_energy:
                dt=temporary_dt
                Rbest=Rnew
                Kbest=Knew
                new_energy=temporary_energy
        R=Rbest
        K=Kbest
        old_energy=new_energy
        print("iteration: "+str(nit) +" energy " + str(new_energy)+ " gradnorm " +str(np.sqrt(np.trace(W @ dagger(W)).real)/D)+ " dt= "+str(dt))
    return R,K
